/*
 * Copyright (C) 2011-2012 by Matthias Ringwald
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. This software may not be used in a commercial product
 *    without an explicit license granted by the copyright holder. 
 *
 * THIS SOFTWARE IS PROVIDED BY MATTHIAS RINGWALD AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MATTHIAS
 * RINGWALD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

//*****************************************************************************
//
// att device demo
//
//*****************************************************************************

// TODO: seperate BR/EDR from LE ACL buffers
// TODO: move LE init into HCI
// ..

// NOTE: Supports only a single connection

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "config.h"
#include "global.h"

#include <btstack/run_loop.h>

#include "btstack_memory.h"
#include "hci.h"
#include "hci_dump.h"
#include "l2cap.h"
#include "btstack/debug.h"

#include "att.h"
// test profile
#include "profile.h"
static uint16_t	sense_lr;		//sibu

int left,right;
#define PI 3.141592
#define neutralAngle 25
#define neutralRange 5
#define angleRange 30.0
#define handlingIntensity 0.7
float AccSin2Deg(uint8_t acc)    // acc : -90 ... 90
{
	return (float)asin((float)(acc-128.0f)/127.0f)*180.0f/PI;
}

float Acc2Speed(uint8_t acc)    // acc : 0 ... 255
{
	int deg = neutralAngle+(int)AccSin2Deg(acc);
	if(deg>-neutralRange && deg<neutralRange)deg=0;
	float speed=100*(float)deg/angleRange;
	if(speed>100){
		speed=100;
	} else if(speed<-100){
		speed=-100;
	}
	return speed;
}

void RCBController(uint8_t *buffer, uint16_t buffer_size)
{
    uint16_t game_pad;
    game_pad = (buffer[0] << 8) | buffer[1];
//  log_info("game_pad : %04X\n",game_pad);
    float AccX = (100.0f/90.0f)*AccSin2Deg(buffer[6]);//((float)buffer[6] / 255.0)*200.0 - 100.0;
//  float AccY = AccSin2Deg(buffer[7]);//((float)buffer[7] / 255.0)*200.0 - 100.0;
    float AccY = Acc2Speed(buffer[7]);
//  float AccZ = (100.0f/90.0f)*AccSin2Deg(buffer[8]);((float)buffer[8] / 255.0)*200.0 - 100.0;
//  log_info("acc X : %f Y : %f Z : %f\n",acc[0],acc[1],acc[2]);
    float LeftStickX  = ((float)buffer[2] / 255.0)*200.0 - 100.0;
    float RightStickX = ((float)buffer[4] / 255.0)*200.0 - 100.0;
    float LeftStickY  = ((float)buffer[3] / 255.0)*200.0 - 100.0;
    float RightStickY = ((float)buffer[5] / 255.0)*200.0 - 100.0;
    uint8_t status = buffer[9];

    if((status & 0x60) == 0x20) {    // Accelerometer ON
        float leftData;
        float rightData;
		float xHandling=(float)AccX*handlingIntensity;
        if(game_pad != 0x0020) {      // A button 
            right = 0;
            left  = 0;            
        } else {
			leftData  = AccY;
			rightData = AccY;
			if(AccY==0) {
				leftData= AccX;
				rightData= -AccX;
			} else if(AccY>0) {
				if(AccX>0) {
					rightData=AccY-(int)xHandling; //r-x
				} else {
					leftData=AccY+(int)xHandling; //l-x
				}
			} else {
				if(AccX>0) {
					leftData=AccY-(int)xHandling; //l-x
				} else {
					rightData=AccY+(int)xHandling; //r-x
				}
			}
            left  = (int)leftData;
            right = (int)rightData;
        }
    } else if((status & 0x10)&&(status & 0x08)) {  // L : Analog R : Analog
        left  = (int)LeftStickY;
        right = (int)RightStickY;
    } else if(status & 0x10) {              // L : Analog
        if((LeftStickX < 15) && (LeftStickX > -15)) {
            right = LeftStickY;
            left  = LeftStickY;            
        }else if((LeftStickY < 15) && (LeftStickY > -15)) {
            right = -LeftStickX;
            left  = LeftStickX;            
        } else {
            right = 0;
            left  = 0;            
        }
    } else if(status & 0x08) {              // R : Analog
        float leftData;
        float rightData;
        if(RightStickY < 0) {
            RightStickX *= -1.0f;
        }
        if(RightStickX >= 0) {
            leftData  = RightStickY + RightStickX;
            rightData = RightStickY;
        } else {
            leftData  = RightStickY;
            rightData = RightStickY - RightStickX;
        }
        if(leftData > 100) {
            leftData = 100;
        } else if(leftData < -100) {
            leftData = -100;
        }
        if(rightData > 100) {
            rightData = 100;
        } else if(rightData < -100) {
            rightData = -100;
        }
        left  = (int)leftData;
        right = (int)rightData;
    } else if((status & 0x60) == 0x40) {    // L : Accelerometer
        if(game_pad != 0x0020) {            // A button 
            right = 0;
            left  = 0;            
        } else {
            if((LeftStickX < 15) && (LeftStickX > -15)) {
                right = LeftStickY;
                left  = LeftStickY;            
            } else if((LeftStickY < 15) && (LeftStickY > -15)) {
                right = -LeftStickX;
                left  = LeftStickX;            
            } else {
                right = 0;
                left  = 0;            
            }
        }
    } else {                        // Digital button
        switch(game_pad)
        {
            case 0x0001:
                left = 50;
                right = 50;
                break;
            case 0x0002:
                left = -50;
                right = -50;
                break;
            case 0x0004:
                left = 50;
                right = -50;
                break;
            case 0x0008:
                left = -50;
                right = 50;
                break;
            default:
                left = 0;
                right = 0;
        }
    }
}

hci_transport_t * hci_transport_picusb_instance();

static att_connection_t att_connection;
static uint16_t         att_response_handle = 0;
static uint16_t         att_response_size   = 0;
static uint8_t          att_response_buffer[28];

uint8_t	connection_status=0;

void hexdump2(void *data, int size){
    int i;
    for (i=0; i<size;i++){
        log_info("%02X ", ((uint8_t *)data)[i]);
    }
    log_info("\n");
}

static void att_try_respond(void){
    if (!att_response_size) return;
    if (!att_response_handle) return;
    if (!hci_can_send_packet_now(HCI_ACL_DATA_PACKET)) return;
    
    // update state before sending packet
    uint16_t size = att_response_size;
    att_response_size = 0;
    l2cap_send_connectionless(att_response_handle, L2CAP_CID_ATTRIBUTE_PROTOCOL, att_response_buffer, size);
}


static void att_packet_handler(uint8_t packet_type, uint16_t handle, uint8_t *packet, uint16_t size){
    if (packet_type != ATT_DATA_PACKET) return;
    
    att_response_handle = handle;
    att_response_size = att_handle_request(&att_connection, packet, size, att_response_buffer);
    att_try_respond();
}


// enable LE, setup ADV data
static void packet_handler (void * connection, uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size){
    static bd_addr_t addr;
//    uint8_t adv_data[] = { 02, 01, 05,   03, 02, 0xf0, 0xff }; 
	const uint8_t adv_data[31]="\x02\x01\x05" "\x08\x09WALLBOT";
    switch (packet_type) {
            
		case HCI_EVENT_PACKET:
			switch (packet[0]) {
				
                case BTSTACK_EVENT_STATE:
					// bt stack activated, get started - set local name
					if (packet[2] == HCI_STATE_WORKING) {
					   log_info("Working!\n");
						hci_send_cmd(&hci_read_local_supported_features);
					}
					break;
                
                case DAEMON_EVENT_HCI_PACKET_SENT:
                    att_try_respond();
                    break;
                    
                case HCI_EVENT_LE_META:
                    switch (packet[2]) {
                        case HCI_SUBEVENT_LE_CONNECTION_COMPLETE:
                            // reset connection MTU
                            att_connection.mtu = 23;
                            break;
                        default:
                            break;
                    }
                    break;

                case BTSTACK_EVENT_NR_CONNECTIONS_CHANGED:
				    if (packet[2]) {
						connection_status=1;
                        log_info("CONNECTED\n");
                    } else {
						connection_status=0;
						sense_lr=0;
						sense_ry=0;
						sense_ly=0;
						out_button_state();

				        log_info("NOT CONNECTED\n");
                    }
                    break;
                    
                case HCI_EVENT_DISCONNECTION_COMPLETE:
                    att_response_handle =0;
                    att_response_size = 0;
                    
                    // restart advertising
                    hci_send_cmd(&hci_le_set_advertise_enable, 1);
                    break;
                    
				case HCI_EVENT_COMMAND_COMPLETE:
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_bd_addr)){
    					bt_flip_addr(addr, &packet[6]);
					    log_info("BD ADDR: %s\n", bd_addr_to_str(addr));
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_local_supported_features)){
					    log_info("Local supported features: %04lX%04lX\n", READ_BT_32(packet, 10), READ_BT_32(packet, 6));
                        hci_send_cmd(&hci_set_event_mask, 0xffffffff, 0x20001fff);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_set_event_mask)){
                        hci_send_cmd(&hci_write_le_host_supported, 1, 1);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_write_le_host_supported)){
				        hci_send_cmd(&hci_le_set_event_mask, 0xffffffff, 0xffffffff);
                    	break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_event_mask)){
				        hci_send_cmd(&hci_le_read_buffer_size);
                    	break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_buffer_size)){
					    log_info("LE buffer size: %u, count %u\n", READ_BT_16(packet,6), packet[8]);
					   hci_send_cmd(&hci_le_read_supported_states);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_supported_states)){
					   hci_send_cmd(&hci_le_set_advertising_parameters,  0x0400, 0x0800, 0, 0, 0, &addr, 0x07, 0);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_parameters)){
					   hci_send_cmd(&hci_le_set_advertising_data, sizeof(adv_data), adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_data)){
					   hci_send_cmd(&hci_le_set_scan_response_data, 10, adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_scan_response_data)){
					   hci_send_cmd(&hci_le_set_advertise_enable, 1);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertise_enable)){
                        hci_discoverable_control(1);
						log_info("startup_state=1\n");
						startup_state=1;
					    led1_off();
                        break;
					}
                    
			}
	}
}

// read requests
static uint16_t att_read_callback(uint16_t handle, uint16_t offset, uint8_t * buffer, uint16_t buffer_size){
	uint16_t	ret=0;

	if(buffer){
		log_info("READ Callback, handle %04x\n", handle);
		led1_on_mac();
	}
    switch(handle){
	case 0x000b:
		if(buffer && buffer_size>=2){
			log_info("Read value: %u\n", sense_lr);
			buffer[0]=sense_lr&0xff;
			buffer[1]=sense_lr>>8;
		}
		ret=2;
		break;
	}
	return ret;
}

// write requests
#define Speed_Low 50
#define Speed_Mid 75
#define Speed_Max 100
static void att_write_callback(uint16_t handle, uint16_t transaction_mode, uint16_t offset, uint8_t *buffer, uint16_t buffer_size, signature_t * signature){
    log_info("WRITE Callback, handle %04x\n", handle);
	led1_on_mac();
    switch(handle){
	case 0x000b:
		sense_lr=(((uint16_t)buffer[0])<<8)|buffer[1];
        RCBController(buffer,buffer_size);
		sense_ry=right;sense_ly=left;
		out_button_state();
		break;
/*		switch (sense_lr)
		{
			case 0x0001: sense_ry= Speed_Mid;	sense_ly= Speed_Mid;   break;
			case 0x0002: sense_ry=-Speed_Mid;   sense_ly=-Speed_Mid;   break;
			case 0x0004: sense_ry=-Speed_Mid;   sense_ly= Speed_Mid;   break;
			case 0x0008: sense_ry= Speed_Mid;   sense_ly=-Speed_Mid;   break;
			case 0x0005: sense_ry= Speed_Mid/2; sense_ly= Speed_Mid;   break;
			case 0x0009: sense_ry= Speed_Mid;   sense_ly= Speed_Mid/2; break;
			case 0x0006: sense_ry=-Speed_Mid/2; sense_ly=-Speed_Mid;   break;
			case 0x000A: sense_ry=-Speed_Mid;   sense_ly=-Speed_Mid/2; break;
			default: sense_ry=0; sense_ly=0; break;
		}
		out_button_state();
		break;
*/
/*	case 0x000b:
		sense_lr=(((uint16_t)buffer[1])<<8)|buffer[0];
		sense_ry=(int)((signed char)buffer[0]);
		sense_ly=(int)((signed char)buffer[1]);
		log_info("sense_lr=%04x\n",sense_lr);
		out_button_state();
		break;
*/   }
}

// main
int blemain(void)
{
	/// GET STARTED with BTstack ///
	btstack_memory_init();
    run_loop_init(RUN_LOOP_EMBEDDED);
	
    // init HCI
	hci_transport_t    * transport = hci_transport_picusb_instance();
	bt_control_t       * control   = NULL;
    hci_uart_config_t  * config    = NULL;
    remote_device_db_t * remote_db = (remote_device_db_t *) &remote_device_db_memory;
	hci_init(transport, config, control, remote_db);
	
    // use eHCILL
    // bt_control_cc256x_enable_ehcill(1);
    
    // set up l2cap_le
    l2cap_init();
    l2cap_register_fixed_channel(att_packet_handler, L2CAP_CID_ATTRIBUTE_PROTOCOL);
    l2cap_register_packet_handler(packet_handler);
    
    // set up ATT
    att_set_db(profile_data);
    att_set_write_callback(att_write_callback);
    att_set_read_callback(att_read_callback);
    att_dump_attributes();
    att_connection.mtu = 27;
    
	log_info("Run...\n\r");

    // turn on!
	hci_power_control(HCI_POWER_ON);

    // go!
    run_loop_execute();	
    
    // happy compiler!
    return 0;
}

